#include <opencv2/opencv.hpp>
#include <fstream>

int saveCalib(std::string fname, cv::Mat &mtx, std::vector<double> &dist, cv::Size &dims)
{
    cv::FileStorage fs(fname, cv::FileStorage::WRITE);
    if (!fs.isOpened()) return -1;
    fs << "mtx" << mtx;
    fs << "dist" << dist;
    fs << "dims" << dims;
    fs.release();
    printf("%s saved", fname.c_str());
    return 0;
}

int loadCalib(std::string fname, cv::Mat &mtx, std::vector<double> &dist, cv::Size &dims)
{
    cv::FileStorage fs(fname, cv::FileStorage::READ);
    if (!fs.isOpened()) return -1;
    fs["mtx"] >> mtx;
    fs["dist"] >> dist;
    fs["dims"] >> dims;
    fs.release();
    printf("%s loaded", fname.c_str());
    return 0;
}
  
int main()
{
    cv::VideoCapture camera(0, cv::CAP_DSHOW);

    int nx = 7; // number of inside corners in x
    int ny = 7; // number of inside corners in y

    std::vector<cv::Point3f> objp;
    for (int y = 0; y < ny; y++)
        for (int x = 0; x < nx; x++)
            objp.emplace_back(float(x), float(y), 0.0f);

    cv::TermCriteria criteria(cv::TermCriteria::EPS | cv::TermCriteria::MAX_ITER, 30, 0.001);

    std::vector<std::vector<cv::Point3f>> objpoints;
    std::vector<std::vector<cv::Point2f>> points;
    cv::Mat mtx, newcameramtx;
    std::vector<double> dist;
    cv::Size dims;

    for (;;) {
        cv::Mat frame, undistorted, gray;

        camera >> frame;
        if (frame.empty()) break;

        cv::imshow("camera", frame);
        if (!mtx.empty()) {
            cv::undistort(frame, undistorted, mtx, dist, newcameramtx);
            cv::imshow("undistorted", undistorted);
        }

        int key = cv::waitKey(10);
        if (key == 27) break;
        else if (key == 'c') {

            // add sample
            cv::cvtColor(frame, gray, cv::COLOR_BGR2GRAY);
            std::vector<cv::Point2f> corners;
            bool found = cv::findChessboardCorners(gray, cv::Size(nx, ny), corners, cv::CALIB_CB_ADAPTIVE_THRESH | cv::CALIB_CB_FAST_CHECK | cv::CALIB_CB_NORMALIZE_IMAGE);
            if (found) {
                dims = cv::Size(gray.cols, gray.rows);

                cv::cornerSubPix(gray, corners, cv::Size(11, 11), cv::Size(-1, -1), criteria);

                cv::drawChessboardCorners(frame, cv::Size(nx, ny), corners, found);
                cv::imshow("chessboardCorners", frame);

                objpoints.push_back(objp);
                points.push_back(corners);

                printf("calibrating\n");
                std::vector<cv::Mat> rvecs, tvecs;
                cv::calibrateCamera(objpoints, points, dims, mtx, dist, rvecs, tvecs);
                cv::Rect roi;
                newcameramtx = cv::getOptimalNewCameraMatrix(mtx, dist, dims, 1, dims, &roi);
                printf("calibrated\n");
            }
            else {
                printf("calibration chessboard not detected\n");
                //break;
            }

        }
        else if (key == 's') {
            if (!mtx.empty()) {
                saveCalib("calib.yaml", mtx, dist, dims);
                printf("calibration saved\n");
            }
        }
        else if (key == 'l') {
            try {
                loadCalib("calib.yaml", mtx, dist, dims);
                cv::Rect roi;
                newcameramtx = cv::getOptimalNewCameraMatrix(mtx, dist, dims, 1, dims, &roi);
                printf("calibration loaded\n");
            }
            catch (...) {}
        }
        else if (key == 'r') {
            objpoints.clear();
            points.clear();
            mtx.release();
            printf("calibration reset\n");
        }
    }

    camera.release();
    cv::destroyAllWindows();
    return 0;
}
